/* natpmp.c:
 * This code is derived from the description provided by the MiniUPnP Project
 * at http://miniupnp.free.fr/nat-pmp.html.
 * Notes (taken from above site):
 *
 * When the public address changes, the NAT gateway must send a notification
 * on the multicast group 224.0.0.1 port 5351 with the format of a public
 * address response. update : in the 2008 version of the specification,
 * port 5350 is used instead.
 *
 * There is no specific request to remove a port mapping. One should use the
 * port mapping request with a lifetime of zero. If lifetime and private port
 * are zero, then all the port mapping to this host are removed.
 *
 * My notes:
 * You'll notice in the code the "useTCP" variable... I initially thought
 * that I had to use TCP to request a TCP port mapping.  WRONG, so this is
 * always zero.  
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>

typedef unsigned char uchar;
typedef unsigned short ushort;

#ifdef BUILD_WITH_VCC
#include <windows.h>

char *optarg;
int optind;
int getopt(int argc,char *argv[],char *opts);

#define uint32_t unsigned long
#define sleep(n) Sleep(n*1000)
#define SOCKET_INVALID(s)	(s == INVALID_SOCKET)

WSADATA WsaData;
DWORD	tid = (DWORD)0;
HANDLE	tHandle;

#else

#include <stdint.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define SOCKET_INVALID(s)	(s < 0)

pthread_attr_t attr;
pthread_t tid = (pthread_t)0;
#endif

#define NATPMP_PORT	 5350 	//5351

struct sockaddr_in targetAddr;

int getResponse(int sockfd, char *msgbuf, int mbsiz, int msecwait, int tcp);

char *errmsg[] = {
	"Usage: natpmp [options] {private port} {public port} [gateway_IP]",
	" Options:",
	"  -s#    port mapping lifetime in seconds (default=1800)",
	"  -t     request is for TCP (default is UDP)",
	"  -v     verbose mode",
	(char *)0
};

void
usage()
{
	int i = 0;

	while(errmsg[i])
		printf("%s\n",errmsg[i++]);
	exit(1);
}

void
err(char *msg)
{
	if (msg)
		fprintf(stderr,"%s, ",msg);
	perror("");
	exit(-1);
}

int
main(int argc,char **argv)
{
	uint32_t inaddr;
	char	*gatewayIP;
	int		i, rc, opt, pktsiz, len, sfd;
	struct	hostent	*hp;
	int verbose = 1;
	int useTCP = 0;		/* never changes, see notes above */
	unsigned char PortMapRqstOpcode = 1;
//	unsigned char PortMapRespOpcode = 129;
	unsigned short private_port;
	unsigned short public_port;
	unsigned long mapping_lifetime = 1800;
	char	natpmprqst[256], natpmpresp[256];

	if (argc == 1)
		usage();
		
	while((opt=getopt(argc,argv,"s:tv")) != EOF) { 
		switch(opt) {
		case 's':
			mapping_lifetime = atoi(optarg);
			break;
		case 'v':
			verbose = 1;
			break;
		case 't':
			PortMapRqstOpcode = 2;
//			PortMapRespOpcode = 130;
			break;
		}
	}

	gatewayIP = "192.168.1.1";

	if (argc == optind+3) {
		gatewayIP = argv[optind+2];
	}
	else if (argc != optind+2)
		usage();
	
	private_port = atoi(argv[optind]);
	public_port = atoi(argv[optind+1]);

	if (verbose) {
		fprintf(stderr,"Private=%d, Public=%d, GW=%s, Lifetime=%ld\n",
			private_port, public_port, gatewayIP, mapping_lifetime);
		fprintf(stderr,"Using %s...\n",useTCP == 1 ? "TCP" : "UDP");
	}

#ifdef BUILD_WITH_VCC
    if (WSAStartup (0x0202, &WsaData) != 0) {
        if (WSAStartup (0x0101, &WsaData) != 0) {
			err("WSAStartup Failed");
        }
    }
#endif

	/* Accept target name as string or internet dotted-decimal address:
	 */
	memset((char *)&targetAddr,0,sizeof(struct sockaddr));
	if ((inaddr = inet_addr(gatewayIP)) != INADDR_NONE) {
		memcpy((char *)&targetAddr.sin_addr,(char *)&inaddr,sizeof(inaddr));
	}
	else {
		hp = gethostbyname(gatewayIP);
		if (hp == NULL)
			err("gethostbyname failed");
		memcpy((char *)&targetAddr.sin_addr,hp->h_addr,hp->h_length);
	}

	if (useTCP)
		sfd = socket(AF_INET,SOCK_STREAM,0);
	else
		sfd = socket(AF_INET,SOCK_DGRAM,0);

	if (SOCKET_INVALID(sfd))
		err("socket failed");

	targetAddr.sin_family = AF_INET;
	targetAddr.sin_port = htons(NATPMP_PORT);

	if (useTCP) {
		rc = connect(sfd,(struct sockaddr *)&targetAddr,
				sizeof(struct sockaddr));
		if (rc < 0) {
			perror("connect");
			exit(1);
		}
	}

	/* Build the NAT-PMP public address request:
	 * ---------------------------
	 * |   BYTE-0   |   BYTE-1   |
	 * ---------------------------
	 * | Version=0  |  OPCode=0  |
	 * ---------------------------
	 */
	natpmprqst[0] = 0x00;	/* Version */
	natpmprqst[1] = 0x00;	/* Opcode */
	
	pktsiz = 2;
	
	if (useTCP) {
		rc = send(sfd,natpmprqst,pktsiz,0);
	}
	else {
		rc = sendto(sfd,natpmprqst,pktsiz,0,
	   	 (struct sockaddr *)&targetAddr,sizeof(targetAddr));
	}

	if (rc < 0) {
		close(sfd);
		err("send/sendto failed");
	}

	/* If a response is received, it should be formatted as follows:
	 * -----------------------------------------------------
	 * |   BYTE-0   |   BYTE-1   |   BYTE-2   |   BYTE-3   |
	 * -----------------------------------------------------
	 * | Version=0  | OPCode=128 |      Result Code        |
	 * -----------------------------------------------------
	 * |   BYTE-4   |   BYTE-5   |   BYTE-6   |   BYTE-7   |
	 * -----------------------------------------------------
	 * | Seconds since port mapping table was initialized  |
	 * -----------------------------------------------------
	 * |   BYTE-8   |   BYTE-9   |   BYTE-10  |   BYTE-11  |
	 * -----------------------------------------------------
	 * |                  Public IP address                |
	 * -----------------------------------------------------
	 */

	len = getResponse(sfd,natpmpresp,sizeof(natpmpresp), 10000, useTCP);
/*
	if (len <= 0)
		exit(1);
*/

	if (verbose)
	{
		printf("Resp: %d bytes\n",len);

		for(i=0;i<len;i++)
			printf("<%02x> %c\n",natpmpresp[i],natpmpresp[i]);
	}

	/* Build the port mapping request.
	 * OPCode = 1 for UDP, 2 for TCP...
	 * -----------------------------------------------------
	 * |   BYTE-0   |   BYTE-1   |   BYTE-2   |   BYTE-3   |
	 * -----------------------------------------------------
	 * | Version=0  | OPCode=1|2 |      Reserved (0)       |
	 * -----------------------------------------------------
	 * |   BYTE-4   |   BYTE-5   |   BYTE-6   |   BYTE-7   |
	 * -----------------------------------------------------
	 * |      Private Port       | Requested public port   |
	 * -----------------------------------------------------
	 * |   BYTE-8   |   BYTE-9   |   BYTE-10  |   BYTE-11  |
	 * -----------------------------------------------------
	 * |    Requested port mapping lifetime in seconds     |
	 * -----------------------------------------------------
	 */

	pktsiz = 12;
	natpmprqst[0] = 0x00;	/* Version */
	natpmprqst[1] = PortMapRqstOpcode;
	natpmprqst[2] = 0x00;	/* Reserved */
	natpmprqst[3] = 0x00;	/* Reserved */
	natpmprqst[4] = (private_port & 0xff00) >> 8;
	natpmprqst[5] = (private_port & 0x00ff);
	natpmprqst[6] = (public_port & 0xff00) >> 8;
	natpmprqst[7] = (public_port & 0x00ff);
	natpmprqst[8] = ((mapping_lifetime & 0xff000000) >> 24);
	natpmprqst[9] = ((mapping_lifetime & 0xff0000) >> 16);
	natpmprqst[10] = ((mapping_lifetime & 0xff00) >> 8);
	natpmprqst[11] = ((mapping_lifetime & 0xff) >> 0);
	
	if (sendto(sfd,natpmprqst,pktsiz,0,
	    (struct sockaddr *)&targetAddr,sizeof(targetAddr)) < 0) {
		close(sfd);
		err("sendto failed");
	}

	/* If a response is received, it should be formatted as follows:
	 * (OPCode is 129 for UDP, 130 for TCP)
	 * -----------------------------------------------------
	 * |   BYTE-0   |   BYTE-1   |   BYTE-2   |   BYTE-3   |
	 * -----------------------------------------------------
	 * | Version=0  |   OPCode   |      Result Code        |
	 * -----------------------------------------------------
	 * |   BYTE-4   |   BYTE-5   |   BYTE-6   |   BYTE-7   |
	 * -----------------------------------------------------
	 * | Seconds since port mapping table was initialized  |
	 * -----------------------------------------------------
	 * |   BYTE-8   |   BYTE-9   |   BYTE-10  |   BYTE-11  |
	 * -----------------------------------------------------
	 * |       Private port      |    Mapped public port   |
	 * -----------------------------------------------------
	 * |   BYTE-12  |   BYTE-13  |   BYTE-14  |   BYTE-15  |
	 * -----------------------------------------------------
	 * |        Port mapping lifetime in seconds           |
	 * -----------------------------------------------------
	 */

	len = getResponse(sfd,natpmpresp,sizeof(natpmpresp), 10000, useTCP);
	if (len <= 0)
		exit(1);

	if (verbose) {
		printf("Resp: %d bytes\n",len);
		for(i=0;i<len;i++)
			printf("<%02x> %c\n",natpmpresp[i], natpmpresp[i]);
	}

	exit(0);
}

int
getResponse(int sockfd, char *msgbuf, int mbsiz, int msecwait, int tcp)
{
	int	msglen, rcvlen;
	fd_set	fd_read;
	struct	timeval tval;

	tval.tv_sec = 0;
	tval.tv_usec = msecwait * 1000;

	FD_ZERO(&fd_read);
	FD_SET(sockfd,&fd_read);

	/* Wait for incoming message: */
	if (select(sockfd+1,&fd_read,0,0,&tval) != 1) {
		return(-1);
	}

	if (tcp) {
		rcvlen = recv(sockfd,msgbuf,mbsiz,0);
	}
	else {
		msglen = sizeof(struct sockaddr);
		rcvlen = recvfrom(sockfd,msgbuf,mbsiz,0,
			(struct sockaddr *)&targetAddr,&msglen);
	}
	if (rcvlen == 0) {
		fprintf(stderr,"Connection closed\n");
	}
	else if (rcvlen < 0) {
		if (tcp)
			perror("recv");
		else
			perror("recvfrom");
	}

	return(rcvlen);
}

#ifdef BUILD_WITH_VCC
/** \var int getopt_sp
 *	\brief Initialized variable used by getopt.
 */
int	getopt_sp = 1;

/** \var int optopt
 *	\brief Global variable used by getopt.
 */
int	optopt;

/** \var int optind
 *	\brief Global variable used by callers to getopt.  It contains
 *	\brief the index into the argument list that is next.
 */
int	optind = 1;

/** \var char *optarg
 *	\brief Pointer to an argument that has an option.
 */
char	*optarg;


/**	\fn int getopt(int argc,char *argv[],char *opts)
 *	\brief Standard mechanism used to retrieve options (dash-prefixed
 *	\brief letters that provide the program with "optional" 
 *	\brief parameters.
 */
int
getopt(int argc,char *argv[],char *opts)
{
	extern char	*strchr();
	register int c;
	register char *cp;

	if (getopt_sp == 1) {
		char	c0, c1;

		if (optind >= argc)
			return(EOF);

		c0 = argv[optind][0];
		c1 = argv[optind][1];

		if (c0 != '-')
			return(EOF);
		if (c1 == '\0')
			return(EOF);
		if (c0 == '-') {
			if ((isdigit(c1)) && (!strchr(opts,c1)))
				return(EOF);
			if (c1 == '-') {
				optind++;
				return(EOF);
			}
		}
	}
	optopt = c = argv[optind][getopt_sp];
	if (c == ':' || ((cp=strchr(opts, c)) == NULL)) {
		fprintf(stderr,"Illegal '%s' option: %c.\n",argv[0],c);
		if(argv[optind][++(getopt_sp)] == '\0') {
			optind++;
			getopt_sp = 1;
		}
		return('?');
	}
	if (*++cp == ':') {
		if(argv[optind][getopt_sp+1] != '\0')
			optarg = &argv[optind++][getopt_sp+1];
		else if(++(optind) >= argc) {
			fprintf(stderr,
				"Option '%c' of '%s' requires argument.\n",
				c,argv[0]);
			getopt_sp = 1;
			return('?');
		} else
			optarg = argv[optind++];
		getopt_sp = 1;
	} else {
		if(argv[optind][++(getopt_sp)] == '\0') {
			getopt_sp = 1;
			optind++;
		}
		optarg = NULL;
	}
	return(c);
}
#endif
